/*
 * glBuild.c
 *
 * Routines to massage a TR2 level into the structures we will use
 * to render the level using OpenGL
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>

#include "config.h"
#include "tr2.h"

#ifdef INTELPC
#ifndef LINUX
/*
 * The #define WIN32_LEAN_AND_MEAN and #include <windows.h> are required for
 * gl.h, glu.h, and glut.h - need to make this more portable...
 */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include "texture.h"
#define OPENGL_FULL                 // tell opengl.h to do all the defines and enums that main.c doesn't want to see
#include "opengl.h"
#include "trmath.h"

/*
 * Macros/Constants
 */
#define ROOM_TRANSPARENCY  0xff  // scale 0 == transparent, 0xFF == opaque

/*
 * Data
 */

/*
 * Externals
 */

/*
 * Code
 */
void opengl_create_all_meshes(void)
{
   int i;

   opengl_free_all_meshes();
   for (i = 0; i < opengl_level->NumRooms; ++i) {
      opengl_set_room_mesh(i);
      }
   for (i = 0; i < opengl_level->NumMeshes; ++i) {
      opengl_set_item_mesh(i);
      }
}

void opengl_free_mesh(opengl_mesh **mesh)
{
   opengl_mesh *meshptr = *mesh;

   if (mesh != NULL && *mesh != NULL) {
      // don't free actual textures;  they're held in an array and can be re-used
      FreeIfUsed(meshptr->oglTexturedRectangleVertices);
      FreeIfUsed(meshptr->oglTexturedRectangleNormals);
      FreeIfUsed(meshptr->oglRectangleTextureCoordinates);
      FreeIfUsed(meshptr->oglRectangleTextures);
      FreeIfUsed(meshptr->oglRectangleVertexColours);

      FreeIfUsed(meshptr->oglTexturedTriangleVertices);
      FreeIfUsed(meshptr->oglTexturedTriangleNormals);
      FreeIfUsed(meshptr->oglTriangleTextureCoordinates);
      FreeIfUsed(meshptr->oglTriangleTextures);
      FreeIfUsed(meshptr->oglTriangleVertexColours);

      FreeIfUsed(meshptr->oglColouredRectangleVertices);
      FreeIfUsed(meshptr->oglColouredRectangleNormals);
      FreeIfUsed(meshptr->oglRectangleColours);
      FreeIfUsed(meshptr->oglCRectangleVertexColours);

      FreeIfUsed(meshptr->oglColouredTriangleVertices);
      FreeIfUsed(meshptr->oglColouredTriangleNormals);
      FreeIfUsed(meshptr->oglTriangleColours);
      FreeIfUsed(meshptr->oglCTriangleVertexColours);

      FreeIfUsed(meshptr);
      *mesh = NULL;
      }
}

void opengl_free_all_meshes(void)
{
   int i;

   for (i = 0; i < MAX_NUM_ROOM_MESHES; ++i)
      opengl_free_mesh(&opengl_item.RoomMeshes[i]);
   for (i = 0; i < MAX_NUM_MESHES; ++i)
      opengl_free_mesh(&opengl_item.ItemMeshes[i]);
}

/*
 * Worker functions for sorting textures with qsort()
 */
int rCompare(const void *element1, const void *element2) // compare rectangles
{
   const tr2_face4 *e1 = element1,
                   *e2 = element2;

   return((int)opengl_level->ObjectTextures[e1->Texture].Tile - (int)opengl_level->ObjectTextures[e2->Texture].Tile);
}

int tCompare(const void *element1, const void *element2) // compare triangles
{
   const tr2_face3 *e1 = element1,
                   *e2 = element2;

   return((int)opengl_level->ObjectTextures[e1->Texture].Tile - (int)opengl_level->ObjectTextures[e2->Texture].Tile);
}

/*
 * Sort a list of textured rectangles by their texture values
 */
void SortRectangles(tr2_face4 *RectangleList, int length)
{
   qsort(RectangleList, length, sizeof(tr2_face4), &rCompare);
}

/*
 * Sort a list of textured triangles by their texture values
 */
void SortTriangles(tr2_face3 *TriangleList, int length)
{
   qsort(TriangleList, length, sizeof(tr2_face3), &tCompare);
}

double AdjustTextureCoordinate(bitu8 pixel, bit8 offset)
{
   if (offset >= 0)
      ++pixel;
   else
      --pixel;
   return((double)pixel / 255.0);
}

/*
 * Initialise an OpenGL (GLfloat) structure for a room mesh.
 */
void opengl_set_room_mesh(int which)
{
   // sort textured rectangles/triangles by texture
   SortRectangles(opengl_level->Rooms[which].RoomData.Rectangles, opengl_level->Rooms[which].RoomData.NumRectangles);
   SortTriangles(opengl_level->Rooms[which].RoomData.Triangles, opengl_level->Rooms[which].RoomData.NumTriangles);
   /* set item struct */
   opengl_item.RoomMeshes[which] = opengl_create_room_mesh(which);
   opengl_item.RoomMeshes[which]->type = ROOM_MESH;
   opengl_item.RoomMeshes[which]->Position.x = opengl_level->Rooms[which].info.x;
   opengl_item.RoomMeshes[which]->Position.z = opengl_level->Rooms[which].info.z;
   opengl_item.RoomMeshes[which]->Position.y = 0;
}

/*
 * Create the actual OpenGL (GLfloat) room mesh structure
 */
opengl_mesh *opengl_create_room_mesh(int index)
{
   opengl_mesh *mesh;
   int i,
       j;
   int v_index;

   /* alloc space for mesh */
   mesh = TR2alloc(sizeof(opengl_mesh));

   mesh->BoundingBoxLow.x = 9999999;
   mesh->BoundingBoxLow.y = 9999999;
   mesh->BoundingBoxLow.z = 9999999;
   mesh->BoundingBoxHigh.x = -9999999;
   mesh->BoundingBoxHigh.y = -9999999;
   mesh->BoundingBoxHigh.z = -9999999;
   // figure out the boundaries of each room (for display when the camera is within)
   for (i = 0; i < opengl_level->Rooms[index].RoomData.NumVertices; ++i) {
      GLfloat q;

      q = opengl_level->Rooms[index].RoomData.Vertices[i].Vertex.x;
      if (q < mesh->BoundingBoxLow.x)
         mesh->BoundingBoxLow.x = q;
      if (q > mesh->BoundingBoxHigh.x)
         mesh->BoundingBoxHigh.x = q;
      q = opengl_level->Rooms[index].RoomData.Vertices[i].Vertex.y;
      if (q < mesh->BoundingBoxLow.y)
         mesh->BoundingBoxLow.y = q;
      if (q > mesh->BoundingBoxHigh.y)
         mesh->BoundingBoxHigh.y = q;
      q = opengl_level->Rooms[index].RoomData.Vertices[i].Vertex.z;
      if (q < mesh->BoundingBoxLow.z)
         mesh->BoundingBoxLow.z = q;
      if (q > mesh->BoundingBoxHigh.z)
         mesh->BoundingBoxHigh.z = q;
      }
   mesh->BoundingBoxLow.x += (GLfloat)(opengl_level->Rooms[index].info.x);
   mesh->BoundingBoxLow.z += (GLfloat)(opengl_level->Rooms[index].info.z);
   mesh->BoundingBoxHigh.x += (GLfloat)(opengl_level->Rooms[index].info.x);
   mesh->BoundingBoxHigh.z += (GLfloat)(opengl_level->Rooms[index].info.z);
   mesh->NumAdjacentRooms = opengl_level->Rooms[index].NumPortals + 1;
   mesh->AdjacentRooms[0] = index;   // current room is always first
   for (i = 0; i < opengl_level->Rooms[index].NumPortals; ++i) {
      mesh->AdjacentRooms[i + 1] = opengl_level->Rooms[index].Portals[i].AdjoiningRoom;
      }

   // for per-vertex lighting
   mesh->oglVertexColours = TR2alloc(sizeof(opengl_colour) * opengl_level->Rooms[index].RoomData.NumVertices);
   for (i = 0; i < opengl_level->Rooms[index].RoomData.NumVertices; ++i) {
      GLfloat ColourValue;
      if (opengl_level->EngineVersion == TombRaider_1)
         ColourValue = opengl_level->Rooms[index].RoomData.Vertices[i].Attributes;
      else
         ColourValue = opengl_level->Rooms[index].RoomData.Vertices[i].Lighting1;
      if (opengl_level->EngineVersion >= TombRaider_3)
         ColourValue /= 16384.0; // was 32768.0; (brighten things up)
      else
         ColourValue = (1.1 - (ColourValue / 8192.0));      // brighten things up?
      if (opengl_level->EngineVersion >= TombRaider_3) {
         #define ROOM_LIGHT_COLOUR_FACTOR 50.0
         mesh->oglVertexColours[i].r = ColourValue + ((GLfloat)opengl_level->Rooms[index].RoomLightColour.Red / ROOM_LIGHT_COLOUR_FACTOR);
         mesh->oglVertexColours[i].g = ColourValue + ((GLfloat)opengl_level->Rooms[index].RoomLightColour.Green / ROOM_LIGHT_COLOUR_FACTOR);
         mesh->oglVertexColours[i].b = ColourValue + ((GLfloat)opengl_level->Rooms[index].RoomLightColour.Blue / ROOM_LIGHT_COLOUR_FACTOR);
         }
      else {
         mesh->oglVertexColours[i].r = ColourValue;
         mesh->oglVertexColours[i].g = ColourValue;
         mesh->oglVertexColours[i].b = ColourValue;
         }
      mesh->oglVertexColours[i].a = 1.0;
      // if the room is filled with water, make its ambient colour a shade of blue...
      if (opengl_level->Rooms[index].Flags & 0x0001) {
         mesh->oglVertexColours[i].r *= 0.6;
         mesh->oglVertexColours[i].b *= 1.2;
         }
      }

   /* get texture mapped rectangle data */
   mesh->oglNumTexturedRectangles = opengl_level->Rooms[index].RoomData.NumRectangles;
   mesh->oglTexturedRectangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedRectangles * 4);
   mesh->oglTexturedRectangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedRectangles * 4);
   mesh->oglRectangleTextureCoordinates = TR2alloc(sizeof(opengl_texture_coord) * mesh->oglNumTexturedRectangles * 4);
   mesh->oglRectangleTextures = TR2alloc(sizeof(opengl_texture *) * mesh->oglNumTexturedRectangles);

   mesh->oglRectangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumTexturedRectangles * 4);
   for (i = 0; i < mesh->oglNumTexturedRectangles; ++i) {
      mesh->oglRectangleTextures[i] = opengl_add_texture(OBJECT, opengl_level->Rooms[index].RoomData.Rectangles[i].Texture, RECTANGLE);
      for (j = 0; j < 4; ++j) {
         v_index = opengl_level->Rooms[index].RoomData.Rectangles[i].Vertices[j];
         mesh->oglRectangleVertexColours[(i * 4) + j] = mesh->oglVertexColours[v_index];
         mesh->oglTexturedRectangleVertices[(i * 4) + j].x = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.x;
         mesh->oglTexturedRectangleVertices[(i * 4) + j].y = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.y;
         mesh->oglTexturedRectangleVertices[(i * 4) + j].z = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.z;
         mesh->oglRectangleTextureCoordinates[(i * 4) + j].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Rectangles[i].Texture].Vertices[j].Xpixel, opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Rectangles[i].Texture].Vertices[j].Xcoordinate);
         mesh->oglRectangleTextureCoordinates[(i * 4) + j].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Rectangles[i].Texture].Vertices[j].Ypixel, opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Rectangles[i].Texture].Vertices[j].Ycoordinate);
         }
      }

   /* get texture mapped triangle data */
   mesh->oglNumTexturedTriangles = opengl_level->Rooms[index].RoomData.NumTriangles;
   mesh->oglTexturedTriangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedTriangles * 3);
   mesh->oglTexturedTriangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedTriangles * 3);
   mesh->oglTriangleTextureCoordinates = TR2alloc(sizeof(opengl_texture_coord) * mesh->oglNumTexturedTriangles * 3);
   mesh->oglTriangleTextures = TR2alloc(sizeof(opengl_texture *) * mesh->oglNumTexturedTriangles);

   mesh->oglTriangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumTexturedTriangles * 3);
   for (i = 0; i < mesh->oglNumTexturedTriangles; ++i) {
      mesh->oglTriangleTextures[i] = opengl_add_texture(OBJECT, opengl_level->Rooms[index].RoomData.Triangles[i].Texture, TRIANGLE);
      for (j = 0; j < 3; ++j) {
         v_index = opengl_level->Rooms[index].RoomData.Triangles[i].Vertices[j];
         mesh->oglTriangleVertexColours[(i * 3) + j] = mesh->oglVertexColours[v_index];
         mesh->oglTexturedTriangleVertices[(i * 3) + j].x = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.x;
         mesh->oglTexturedTriangleVertices[(i * 3) + j].y = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.y;
         mesh->oglTexturedTriangleVertices[(i * 3) + j].z = opengl_level->Rooms[index].RoomData.Vertices[v_index].Vertex.z;
         mesh->oglTriangleTextureCoordinates[(i * 3) + j].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Triangles[i].Texture].Vertices[j].Xpixel, opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Triangles[i].Texture].Vertices[j].Xcoordinate);
         mesh->oglTriangleTextureCoordinates[(i * 3) + j].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Triangles[i].Texture].Vertices[j].Ypixel, opengl_level->ObjectTextures[opengl_level->Rooms[index].RoomData.Triangles[i].Texture].Vertices[j].Ycoordinate);
         }
      }

   free(mesh->oglVertexColours);

   /* get portal data (as magenta-coloured rectangles) */
   mesh->oglNumColouredRectangles = opengl_level->Rooms[index].NumPortals;
   mesh->oglColouredRectangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredRectangles * 4);
   mesh->oglColouredRectangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredRectangles * 4);
   mesh->oglRectangleColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumColouredRectangles);
   for (i = 0; i < mesh->oglNumColouredRectangles; ++i) {
      for (j = 0; j < 4; ++j) {
         mesh->oglColouredRectangleVertices[(i * 4) + j].x = opengl_level->Rooms[index].Portals[i].Vertices[j].x;
         mesh->oglColouredRectangleVertices[(i * 4) + j].y = opengl_level->Rooms[index].Portals[i].Vertices[j].y;
         mesh->oglColouredRectangleVertices[(i * 4) + j].z = opengl_level->Rooms[index].Portals[i].Vertices[j].z;
         mesh->oglColouredRectangleNormals[(i * 4) + j].x = opengl_level->Rooms[index].Portals[i].Normal.x / NORMAL_SCALE;
         mesh->oglColouredRectangleNormals[(i * 4) + j].y = opengl_level->Rooms[index].Portals[i].Normal.y / NORMAL_SCALE;
         mesh->oglColouredRectangleNormals[(i * 4) + j].z = opengl_level->Rooms[index].Portals[i].Normal.z / NORMAL_SCALE;
         }
      mesh->oglRectangleColours[i].r = 1.0; // make portals magenta (if visible)
      mesh->oglRectangleColours[i].g = 0.0;
      mesh->oglRectangleColours[i].b = 1.0;
      mesh->oglRectangleColours[i].a = 1.0;
      }

   return(mesh);
}

/*
 * Initialise an OpenGL (GLfloat) structure for an item mesh.
 */
void opengl_set_item_mesh(int which)
{
   // sort textured rectangles/triangles by texture (for efficiency when drawing)
   SortRectangles(opengl_level->Meshes[which].TexturedRectangles, opengl_level->Meshes[which].NumTexturedRectangles);
   SortTriangles(opengl_level->Meshes[which].TexturedTriangles, opengl_level->Meshes[which].NumTexturedTriangles);

   /* set item structure */
   opengl_item.ItemMeshes[which] = opengl_create_item_mesh(which);
   opengl_item.ItemMeshes[which]->type = ITEM_MESH;
   opengl_item.ItemMeshes[which]->Position.x = 0;
   opengl_item.ItemMeshes[which]->Position.y = 0;
   opengl_item.ItemMeshes[which]->Position.z = 0;
}

/*
 * Create the actual OpenGL (GLfloat) item mesh structure
 */
opengl_mesh *opengl_create_item_mesh(int index)
{
   opengl_mesh *mesh;
   int i,
       j;
   int v_index;
   int c_index;

   /* alloc space for mesh */
   mesh = TR2alloc(sizeof(opengl_mesh));

   if (opengl_level->Meshes[index].MeshLights != NULL) {
      // for per-vertex lighting
      mesh->oglVertexColours = TR2alloc(sizeof(opengl_colour) * opengl_level->Meshes[index].NumVertices);
      for (i = 0; i < opengl_level->Meshes[index].NumVertices; ++i) {
         GLfloat ColourValue;
         ColourValue = opengl_level->Meshes[index].MeshLights[i];
         if (opengl_level->EngineVersion >= TombRaider_3)
            ColourValue /= 16384.0; // 32768.0; - brighten things up
         else
            ColourValue = (1.0 - (ColourValue / 8192.0));
         mesh->oglVertexColours[i].r = ColourValue;
         mesh->oglVertexColours[i].g = ColourValue;
         mesh->oglVertexColours[i].b = ColourValue;
         mesh->oglVertexColours[i].a = 1.0;
         }
      }

   /* get texture mapped rectangle data */
   mesh->oglNumTexturedRectangles = opengl_level->Meshes[index].NumTexturedRectangles;
   if (opengl_level->Meshes[index].MeshLights != NULL) {
      mesh->oglRectangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumTexturedRectangles * 4);
      }
   else {
      mesh->oglTexturedRectangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedRectangles * 4);
      }
   mesh->oglTexturedRectangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedRectangles * 4);
   mesh->oglRectangleTextureCoordinates = TR2alloc(sizeof(opengl_texture_coord) * mesh->oglNumTexturedRectangles * 4);
   mesh->oglRectangleTextures = TR2alloc(sizeof(opengl_texture *) * mesh->oglNumTexturedRectangles);
   for (i = 0; i < mesh->oglNumTexturedRectangles; ++i) {
      mesh->oglRectangleTextures[i] = opengl_add_texture(OBJECT, opengl_level->Meshes[index].TexturedRectangles[i].Texture, RECTANGLE);
      for (j = 0; j < 4; ++j) {
         v_index = opengl_level->Meshes[index].TexturedRectangles[i].Vertices[j];
         if (opengl_level->Meshes[index].MeshLights != NULL) {
            mesh->oglRectangleVertexColours[(i * 4) + j] = mesh->oglVertexColours[v_index];
            }
         else {
            mesh->oglTexturedRectangleNormals[(i * 4) + j].x = opengl_level->Meshes[index].Normals[v_index].x / NORMAL_SCALE;
            mesh->oglTexturedRectangleNormals[(i * 4) + j].y = opengl_level->Meshes[index].Normals[v_index].y / NORMAL_SCALE;
            mesh->oglTexturedRectangleNormals[(i * 4) + j].z = opengl_level->Meshes[index].Normals[v_index].z / NORMAL_SCALE;
            }
         mesh->oglTexturedRectangleVertices[(i * 4) + j].x = opengl_level->Meshes[index].Vertices[v_index].x;
         mesh->oglTexturedRectangleVertices[(i * 4) + j].y = opengl_level->Meshes[index].Vertices[v_index].y;
         mesh->oglTexturedRectangleVertices[(i * 4) + j].z = opengl_level->Meshes[index].Vertices[v_index].z;
         mesh->oglRectangleTextureCoordinates[(i * 4) + j].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedRectangles[i].Texture].Vertices[j].Xpixel, opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedRectangles[i].Texture].Vertices[j].Xcoordinate);
         mesh->oglRectangleTextureCoordinates[(i * 4) + j].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedRectangles[i].Texture].Vertices[j].Ypixel, opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedRectangles[i].Texture].Vertices[j].Ycoordinate);
         }
      }
   if (opengl_level->EngineVersion == TombRaider_3) { // try to account for TR3 alpha
      for (i = 0; i < mesh->oglNumTexturedRectangles; ++i) {
         if (opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedRectangles[i].Texture].TransparencyFlags != 2)
            break;
         }
      if (i == mesh->oglNumTexturedRectangles) {
         FreeIfUsed(mesh->oglRectangleVertexColours);
         mesh->oglRectangleVertexColours = NULL;
         }
      }

   /* get texture mapped triangle data */
   mesh->oglNumTexturedTriangles = opengl_level->Meshes[index].NumTexturedTriangles;
   if (opengl_level->Meshes[index].MeshLights != NULL) {
      mesh->oglTriangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumTexturedTriangles * 3);
      }
   else {
      mesh->oglTexturedTriangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedTriangles * 3);
      }
   mesh->oglTexturedTriangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumTexturedTriangles * 3);
   mesh->oglTriangleTextureCoordinates = TR2alloc(sizeof(opengl_texture_coord) * mesh->oglNumTexturedTriangles * 3);
   mesh->oglTriangleTextures = TR2alloc(sizeof(opengl_texture *) * mesh->oglNumTexturedTriangles);
   for (i = 0; i < mesh->oglNumTexturedTriangles; ++i) {
      mesh->oglTriangleTextures[i] = opengl_add_texture(OBJECT, opengl_level->Meshes[index].TexturedTriangles[i].Texture, TRIANGLE);
      for (j = 0; j < 3; ++j) {
         v_index = opengl_level->Meshes[index].TexturedTriangles[i].Vertices[j];
         if (opengl_level->Meshes[index].MeshLights != NULL) {
            mesh->oglTriangleVertexColours[(i * 3) + j] = mesh->oglVertexColours[v_index];
            }
         else {
            mesh->oglTexturedTriangleNormals[(i * 3) + j].x = opengl_level->Meshes[index].Normals[v_index].x / NORMAL_SCALE;
            mesh->oglTexturedTriangleNormals[(i * 3) + j].y = opengl_level->Meshes[index].Normals[v_index].y / NORMAL_SCALE;
            mesh->oglTexturedTriangleNormals[(i * 3) + j].z = opengl_level->Meshes[index].Normals[v_index].z / NORMAL_SCALE;
            }
         mesh->oglTexturedTriangleVertices[(i * 3) + j].x = opengl_level->Meshes[index].Vertices[v_index].x;
         mesh->oglTexturedTriangleVertices[(i * 3) + j].y = opengl_level->Meshes[index].Vertices[v_index].y;
         mesh->oglTexturedTriangleVertices[(i * 3) + j].z = opengl_level->Meshes[index].Vertices[v_index].z;
         mesh->oglTriangleTextureCoordinates[(i * 3) + j].s = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedTriangles[i].Texture].Vertices[j].Xpixel, opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedTriangles[i].Texture].Vertices[j].Xcoordinate);
         mesh->oglTriangleTextureCoordinates[(i * 3) + j].t = AdjustTextureCoordinate(opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedTriangles[i].Texture].Vertices[j].Ypixel, opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedTriangles[i].Texture].Vertices[j].Ycoordinate);
         }
      }
   if (opengl_level->EngineVersion == TombRaider_3) { // try to account for TR3 alpha
      for (i = 0; i < mesh->oglNumTexturedTriangles; ++i) {
         if (opengl_level->ObjectTextures[opengl_level->Meshes[index].TexturedTriangles[i].Texture].TransparencyFlags != 2)
            break;
         }
      if (i == mesh->oglNumTexturedTriangles) {
         FreeIfUsed(mesh->oglTriangleVertexColours);
         mesh->oglTriangleVertexColours = NULL;
         }
      }

   /* get coloured rectangle data */
   mesh->oglNumColouredRectangles = opengl_level->Meshes[index].NumColouredRectangles;
   if (mesh->oglNumColouredRectangles != 0) {
      if (opengl_level->Meshes[index].MeshLights != NULL) {
         mesh->oglCRectangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumColouredRectangles * 4);
         }
      else {
         mesh->oglColouredRectangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredRectangles * 4);
         }
      mesh->oglColouredRectangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredRectangles * 4);
      mesh->oglRectangleColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumColouredRectangles);
      for (i = 0; i < mesh->oglNumColouredRectangles; ++i) {
         for (j = 0; j < 4; ++j) {
            v_index = opengl_level->Meshes[index].ColouredRectangles[i].Vertices[j];
            if (opengl_level->Meshes[index].MeshLights != NULL) {
               mesh->oglCRectangleVertexColours[(i * 4) + j] = mesh->oglVertexColours[v_index];
               }
            else {
               mesh->oglColouredRectangleNormals[(i * 4) + j].x = opengl_level->Meshes[index].Normals[v_index].x / NORMAL_SCALE;
               mesh->oglColouredRectangleNormals[(i * 4) + j].y = opengl_level->Meshes[index].Normals[v_index].y / NORMAL_SCALE;
               mesh->oglColouredRectangleNormals[(i * 4) + j].z = opengl_level->Meshes[index].Normals[v_index].z / NORMAL_SCALE;
               }
            mesh->oglColouredRectangleVertices[(i * 4) + j].x = opengl_level->Meshes[index].Vertices[v_index].x;
            mesh->oglColouredRectangleVertices[(i * 4) + j].y = opengl_level->Meshes[index].Vertices[v_index].y;
            mesh->oglColouredRectangleVertices[(i * 4) + j].z = opengl_level->Meshes[index].Vertices[v_index].z;
            }
         if (opengl_level->EngineVersion == TombRaider_1) { // no 16-bit palette here...
            c_index = opengl_level->Meshes[index].ColouredRectangles[i].Texture & 0xff;
            mesh->oglRectangleColours[i].r = (double)opengl_level->Palette8[c_index].Red / TR2_COLOUR_MAX;
            mesh->oglRectangleColours[i].g = (double)opengl_level->Palette8[c_index].Green / TR2_COLOUR_MAX;
            mesh->oglRectangleColours[i].b = (double)opengl_level->Palette8[c_index].Blue / TR2_COLOUR_MAX;
            mesh->oglRectangleColours[i].a = 1.0;
            }
         else {
            // NOTE: this assumes that Palette16[] was not endian-converted...
            c_index = (opengl_level->Meshes[index].ColouredRectangles[i].Texture >> 8) & 0xff;
            mesh->oglRectangleColours[i].r = (double)(opengl_level->Palette16[c_index] & 0xff) / 256.0;
            mesh->oglRectangleColours[i].g = (double)((opengl_level->Palette16[c_index] >> 8) & 0xff) / 256.0;
            mesh->oglRectangleColours[i].b = (double)((opengl_level->Palette16[c_index] >> 16) & 0xff) / 256.0;
            mesh->oglRectangleColours[i].a = 1.0;
            }
         }
      }

   /* get coloured triangle data */
   mesh->oglNumColouredTriangles = opengl_level->Meshes[index].NumColouredTriangles;
   if (mesh->oglNumColouredTriangles != 0) {
      if (opengl_level->Meshes[index].MeshLights != NULL) {
         mesh->oglCTriangleVertexColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumColouredTriangles * 3);
         }
      else {
         mesh->oglColouredTriangleNormals = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredTriangles * 3);
         }
      mesh->oglColouredTriangleVertices = TR2alloc(sizeof(opengl_vector) * mesh->oglNumColouredTriangles * 3);
      mesh->oglTriangleColours = TR2alloc(sizeof(opengl_colour) * mesh->oglNumColouredTriangles);
      for (i = 0; i < mesh->oglNumColouredTriangles; ++i) {
         for (j = 0; j < 3; ++j) {
            v_index = opengl_level->Meshes[index].ColouredTriangles[i].Vertices[j];
            if (opengl_level->Meshes[index].MeshLights != NULL) {
               mesh->oglCTriangleVertexColours[(i * 3) + j] = mesh->oglVertexColours[v_index];
               }
            else {
               mesh->oglColouredTriangleNormals[(i * 3) + j].x = opengl_level->Meshes[index].Normals[v_index].x / NORMAL_SCALE;
               mesh->oglColouredTriangleNormals[(i * 3) + j].y = opengl_level->Meshes[index].Normals[v_index].y / NORMAL_SCALE;
               mesh->oglColouredTriangleNormals[(i * 3) + j].z = opengl_level->Meshes[index].Normals[v_index].z / NORMAL_SCALE;
               }
            mesh->oglColouredTriangleVertices[(i * 3) + j].x = opengl_level->Meshes[index].Vertices[v_index].x;
            mesh->oglColouredTriangleVertices[(i * 3) + j].y = opengl_level->Meshes[index].Vertices[v_index].y;
            mesh->oglColouredTriangleVertices[(i * 3) + j].z = opengl_level->Meshes[index].Vertices[v_index].z;
            }
         if (opengl_level->EngineVersion == TombRaider_1) { // no 16-bit palette here...
            c_index = opengl_level->Meshes[index].ColouredTriangles[i].Texture & 0xff;
            mesh->oglTriangleColours[i].r = (double)opengl_level->Palette8[c_index].Red / TR2_COLOUR_MAX;
            mesh->oglTriangleColours[i].g = (double)opengl_level->Palette8[c_index].Green / TR2_COLOUR_MAX;
            mesh->oglTriangleColours[i].b = (double)opengl_level->Palette8[c_index].Blue / TR2_COLOUR_MAX;
            mesh->oglTriangleColours[i].a = 1.0;
            }
         else {
            // NOTE: this assumes that Palette16[] was not endian-converted...
            c_index = (opengl_level->Meshes[index].ColouredTriangles[i].Texture >> 8) & 0xff;
            mesh->oglTriangleColours[i].r = (double)(opengl_level->Palette16[c_index] & 0xff) / 256.0;
            mesh->oglTriangleColours[i].g = (double)((opengl_level->Palette16[c_index] >> 8) & 0xff) / 256.0;
            mesh->oglTriangleColours[i].b = (double)((opengl_level->Palette16[c_index] >> 16) & 0xff) / 256.0;
            mesh->oglTriangleColours[i].a = 1.0;
            }
         }
      }

   FreeIfUsed(mesh->oglVertexColours);

   return(mesh);
}


