
/*
 * Tomb Raider II Data File Format
 *
 * Last Modified: 12feb00
 */
#include "tr2port.h" // portability header

#ifndef TR2_H     // prevent accidental double-inclusion errors
#define TR2_H  1

/*
 * Some useful makros
 */

#define dimof(a) (sizeof(a)/sizeof(a[0]))
#define dim2of(a) (sizeof(a)/sizeof(a[0][0]))

/*
 * For testing only
 */

#ifdef DEBUG
#define PRINTF(x) printf x
#else
#define PRINTF(x) /*nop*/
#endif

/*
 * Miscellaneous constants
 */
#undef TRUE             // force TRUE/FALSE to be 1/0 (just in case)
#define TRUE   1
#undef FALSE
#define FALSE  0
#ifndef __cplusplus
#define true   (TRUE)   // for the C++ folks
#define false  (FALSE)  // ditto
#endif // __cplusplus

/*
 * Version info (for TR1/2/3)
 */
typedef enum {
   TombRaider_UnknownVersion,
   TombRaider_1,
   TombRaider_2,
   TombRaider_3,
   TombRaider_4
   } TR2_Version_type;

/*
 * File or buffer
 */

typedef struct tr2_file_struct {
   TR2_Version_type Version;
   FILE             *Fp;
   bitu8            *Data;
   bitu32           DataSize;
   bitu32           DataOffset;
   } tr2_file;

/*
 * Basic 24-bit colour structure (although it appears that only 6 bits
 * per colour are actually used, making it 18-bit colour)
 */
typedef struct tr2_colour_struct {  // 3 bytes
   bitu8 Red;
   bitu8 Green;
   bitu8 Blue;
   } tr2_colour;

/*
 * Basic vertex structure.
 * Note that all vertices are relative coordinates;
 * each mesh (room, object, etc.) has its own offset by
 * which the vertex coordinates are translated.
 */
typedef struct tr2_vertex_struct {  // 6 bytes
   bit16 x;
   bit16 y;
   bit16 z;
   } tr2_vertex;

/*
 * A rectangular (quad) face definition.  Four vertices (the values are
 * indices into the appropriate vertex list) and a texture (an index
 * into the texture list) or colour (index into 8-bit palette).
 * I've seen a few coloured polygons where Texture is greater than 255,
 * but ANDing the value with 0xFF seems to produce acceptable results.
 */
typedef struct tr2_face4_struct {  // 10 bytes
   bitu16 Vertices[4];
   bitu16 Texture;
   } tr2_face4;

/*
 * A triangular face definition.  Three vertices (the values are indices
 * into the appropriate vertex list) and a texture (an index into the
 * texture list) or colour (index into palette).
 * In the case of a colour, (Texture & 0xff) is the index into the 8-bit
 * palette, while (Texture >> 8) is the index into the 16-bit palette.
 */
typedef struct tr2_face3_struct {  // 8 bytes
   bitu16 Vertices[3];
   bitu16 Texture;
   } tr2_face3;

/*
 * An 8-bit texture tile (65536 bytes).  Each byte represents a pixel
 * whose colour is in the 8-bit palette.
 */
typedef struct tr2_textile8_struct {   // 65536 bytes
   bitu8 Tile[256*256];
   } tr2_textile8;

/*
 * A 16-bit texture tile (131072 bytes).  Each word represents a pixel
 * whose colour is of the form ARGB, MSB-to-LSB:
 *    1-bit transparency (0: transparent, 1: opaque)
 *    5-bit red channel
 *    5-bit green channel
 *    5-bit blue channel
 */
typedef struct tr2_textile16_struct {  // 131072 bytes
   bitu16 Tile[256*256];
   } tr2_textile16;

/*
 * The "header" of a room.  x/z indicate the base position of the room
 * mesh in world coordinates.  yLowest and yHighest indicate the lowest and
 * highest points in this room (even though their actual values appear to
 * be reversed, since a "high" point will have a smaller value than a "low"
 * point).  When positioning objects/items, as well as the room meshes
 * themselves, y is always 0-relative (not room-relative).
 */
typedef struct tr2_room_info_struct {  // 16 bytes
   bit32 x;             // X-offset of room (world coordinates)
   bit32 z;             // Z-offset of room (world coordinates)
   bit32 yBottom;       // Y-offset of lowest point in room (world coordinates) (actually highest value)
   bit32 yTop;          // Y-offset of highest point in room (world coordinates) (actually lowest value)
   } tr2_room_info;

/*
 * Portal structure.  This defines every viable exit from a given "room".
 * Note that "rooms" are really just areas;  they aren't necessarily enclosed.
 * The door structure below defines areas of egress, not the actual moveable
 * mesh, texture, and action (if any).
 */
typedef struct tr2_room_portal_struct {   // 32 bytes
   bitu16 AdjoiningRoom;                  // which room this "door" leads to
   struct tr2_vertex_struct Normal;       // which way the "door" faces
   struct tr2_vertex_struct Vertices[4];  // the corners of the "door"
   } tr2_room_portal;

/*
 * Room sector structure.
 *
 * Sectors are 1024 * 1024 (world coordinates).  Floor and Ceiling are signed
 * number of 256 units of height (relative to 0), e.g. Floor 0x04 corresponds to
 * Y = 1024 in world coordinates.  Note: this implies that, while X and Z can
 * be quite large, Y is constrained to -32768..32512.  Floor/Ceiling value of
 * 0x81 indicates impenetrable wall.
 * Floor values are used by the game engine to determine what objects Lara can
 * traverse and how.  Relative steps of 1 (-256) can be walked up;  steps of
 * 2..7 (-512..-1792) can/must be climbed;  steps larger than 7 (-2048..-32768)
 * cannot be climbed (too tall).
 */
typedef struct tr2_room_sector_struct {   // 8 bytes
   bitu16 FDindex;
   bitu16 BoxIndex;
   bitu8  RoomBelow;             // 255 if none
   bit8   Floor;
   bitu8  RoomAbove;             // 255 if none
   bit8   Ceiling;
   } tr2_room_sector;

/*
 * Room lighting structure.  X/Y/Z are in world coordinates.
 * Lighting values seem to range from 0..8192.
 */
typedef struct tr2_room_light_struct { // 24 bytes
   bit32  x,
          y,
          z;
   bitu16 Intensity1,
          Intensity2;
   bitu32 Fade1,
          Fade2;
   bitu8  junk[22];     // TR4 only
   } tr2_room_light;

/*
 * Room vertex structure.  This defines the vertices within a room.
 */
typedef struct tr2_vertex_room_struct {   // 12 bytes
   tr2_vertex Vertex;
                        // following 3 entries
   bit16 Lighting1;     // values range from 0 to 32767, 0=total darkness. (TR3)
                        // I think the values ranged from 0 to 8192 in TR1/2, 0=total brightness
   bitu16 Attributes;   // 0x8000 something to do with water surface
                        // 0x4000 under water lighting modulation and movement if viewed from above water surface
                        // 0x2000 water/quicksand surface movement
                        // 0x1fef nothing?
                        // 0x0010 everything?
   bit16 Lighting2;     // seems to be the same as lighting1
   } tr2_vertex_room;

/*
 * Sprite structure
 */
typedef struct tr2_room_sprite_struct {   // 4 bytes
   bit16 Vertex;     // offset into vertex list
   bit16 Texture;    // offset into texture list
   } tr2_room_sprite;

/*
 * Room mesh structure.  This is the geometry of the "room," including
 * walls, floors, rocks, water, etc.  It does _not_ include objects that
 * Lara can interact with (keyboxes, moveable blocks, moveable doors, etc.)
 */
typedef struct tr2_room_data_struct {
   bit16           NumVertices;     // number of vertices in the following list
   tr2_vertex_room *Vertices;       // list of vertices (relative coordinates)
   bit16           NumRectangles;   // number of textured rectangles
   tr2_face4       *Rectangles;     // list of textured rectangles
   bit16           NumTriangles;    // number of textured triangles
   tr2_face3       *Triangles;      // list of textured triangles
   bit16           NumSprites;      // number of sprites
   tr2_room_sprite *Sprites;        // list of sprites
   } tr2_room_data;

/*
 * Room static mesh data.  Positions and IDs of static meshes (e.g. skeletons,
 * spiderwebs, furniture)
 */
typedef struct tr2_room_staticmesh_struct {  // 20 bytes
   bit32  x,      // absolute position in world coordinates
          y,
          z;
   bitu16 Rotation;  // high two bits (0xc000) indicate steps of 90 degrees
   bitu16 Intensity1,
          Intensity2;
   bitu16 ObjectID;     // which StaticMesh item to draw
   } tr2_room_staticmesh;

/*
 * Room structure.  Here's where all the room data come together.
 */
typedef struct tr2_room_struct {
   struct tr2_room_info_struct info;            // where the room exists, in world coordinates
   bitu32 NumDataWords;                         // number of data words (bitu16)
   bitu8  *Data;                                // the raw data from which the rest of this is derived
   struct tr2_room_data_struct RoomData;        // the room mesh
   bitu16 NumPortals;                           // number of visibility portals that leave this room
   struct tr2_room_portal_struct *Portals;      // list of visibility portals
   bitu16 NumZsectors;                          // width of sector list
   bitu16 NumXsectors;                          // height of sector list
   struct tr2_room_sector_struct *SectorList;   // list of sectors in this room
   bit16  Intensity1,
          Intensity2,
          LightMode;
   bitu16 NumLights;                            // number of lights in this room
   struct tr2_room_light_struct *Lights;        // list of lights
   bitu16 NumStaticMeshes;                      // number of static meshes
   tr2_room_staticmesh *StaticMeshes;           // static meshes
   bit16  AlternateRoom;
   bit16  Flags;                                // 0x0001 - room is filled with water
                                                // 0x0020 - Lara's ponytail gets blown by the wind
   tr2_colour RoomLightColour;                  // TR3 ONLY!
   } tr2_room;

/*
 * Animation structure.
 */
typedef struct tr2_animation_struct { // 32 bytes
   bitu32 FrameOffset;     // byte offset into Frames[] (divide by 2 for Frames[i])
   bitu8  FrameRate;       // "ticks" per frame
   bitu8  FrameSize;       // number of words in Frames[] used by this animation
   bit16  StateID,
          Unknown1,
          Unknown2,
          Unknown3,
          Unknown4;
   bitu8  junk[8];         // TR4 only
   bitu16 FrameStart;      // first frame in this animation
   bitu16 FrameEnd;        // last frame in this animation (numframes = (End - Start) + 1)
   bitu16 NextAnimation,
          NextFrame,
          NumStateChanges;
   bitu16 StateChangeOffset;  // offset into StateChanges[]
   bitu16 NumAnimCommands;
   bitu16 AnimCommand;        // offset into AnimCommands[]
   } tr2_animation;

/*
 * State Change structure
 */
typedef struct tr2_state_change_struct {  // 6 bytes
   bitu16 StateID;
   bitu16 NumAnimDispatches;  // number of dispatches (seems to always be 1..5)
   bitu16 AnimDispatch;       // Offset into AnimDispatches[]
   } tr2_state_change;

/*
 * Animation Dispatch structure
 */
typedef struct tr2_anim_dispatch_struct {   // 8 bytes
   bit16 Low,
         High,
         NextAnimation,
         NextFrame;
   } tr2_anim_dispatch;

/*
 * AnimCommand structure
 */
typedef struct tr2_anim_command_struct {   // 2 bytes
   bit16 Value;
   } tr2_anim_command;

/*
 * MeshTree structure
 *
 * MeshTree[] is actually groups of four bit32s.  The first one is a
 * "flags" word;  bit 1 (0x0002) indicates "make previous mesh an
 * anchor (e.g. PUSH)";  bit 0 (0x0001) indicates "return to previous
 * anchor (e.g. POP)".
 * The next three bit32s are X, Y, Z offsets from the last mesh position.
 */
typedef struct tr2_meshtree_struct {   // 16 bytes
   bit32 Flags;   // 0x0001 = POP, 0x0002 = PUSH
   bit32 x,
         y,
         z;
   } tr2_meshtree;

/*
 * Moveable structure.  This defines a list of contiguous meshes that
 * comprise one object, e.g. in WALL.TR2,
 *    moveable[0]  = Lara (StartingMesh 0, NumMeshes 15),
 *    moveable[13] = Tiger (StartingMesh 215, NumMeshes 27)
 *    moveable[15] = Spikes (StartingMesh 249, NumMeshes 1)
 *    moveable[16] = Boulder (StartingMesh 250, NumMeshes 1)
 *    moveable[20] = Rolling Blade (StartingMesh 254, NumMeshes 1)
 */
typedef struct tr2_moveable_struct {   // 18 bytes
   bitu32 ObjectID;        // Item Identifier
   bitu16 NumMeshes;       // number of meshes in this object
   bitu16 StartingMesh;    // first mesh
   bitu32 MeshTree;        // offset into MeshTree[]
   bitu32 FrameOffset;     // byte offset into Frames[] (divide by 2 for Frames[i])
   bitu16 Animation;       // offset into Animations[]
   } tr2_moveable;

/*
 * StaticMesh structure.  This defines meshes that don't move (e.g. skeletons
 * lying on the floor, spiderwebs, etc.)
 */
typedef struct tr2_staticmesh_struct { // 32 bytes
   bitu32 ObjectID;           // Item Identifier
   bitu16 StartingMesh;       // first mesh
   tr2_vertex BoundingBox[2][2];
   bitu16 Flags;
   } tr2_staticmesh;

/*
 * Object texture vertex structure.  Maps coordinates
 * into texture tiles.
 */
typedef struct tr2_object_texture_vert_struct {  // 4 bytes
   bitu8 Xcoordinate;
   bitu8 Xpixel;
   bitu8 Ycoordinate;
   bitu8 Ypixel;
   } tr2_object_texture_vert;

/*
 * Object texture structure.
 */
typedef struct tr2_object_texture_struct {    // 20 bytes
   bitu16 TransparencyFlags;  // (0: Opaque; 1: Use transparency; 2: Use partial transparency [grayscale intensity :: transparency])
   bitu16 Tile;            // index into textile list
   struct tr2_object_texture_vert_struct Vertices[4];   // the four corners of the texture
   bitu8  filler[16];  // TR4 only
   } tr2_object_texture;

/*
 * Sprite texture structure.
 */
typedef struct tr2_sprite_texture_struct { // 16 bytes
   bitu16 Tile;
   bitu8  x;
   bitu8  y;
   bitu16 Width;     // actually, (width * 256) + 255
   bitu16 Height;    // actually, (height * 256) + 255
   bit16  LeftSide;
   bit16  TopSide;
   bit16  RightSide;
   bit16  BottomSide;
   } tr2_sprite_texture;

/*
 * Sprite Sequence structure
 */
typedef struct tr2_sprite_sequence_struct {  // 8 bytes
   bit32 ObjectID;         // Item identifier (same numbering as in tr2_moveable)
   bit16 NegativeLength;   // negative of "how many sprites are in this sequence"
   bit16 Offset;           // where (in sprite texture list) this sequence starts
   } tr2_sprite_sequence;

/*
 * Mesh structure.  The mesh list contains the mesh info for Lara (in all her
 * various incarnations), blocks, enemies (tigers, birds, bad guys), moveable
 * blocks, zip line handles, boulders, spinning blades, you name it.
 * If NumNormals is negative, Normals[] represent vertex lighting values (one
 * per vertex).
 */
typedef struct tr2_mesh_struct {
   tr2_vertex Centre;                  // this seems to describe the approximate geometric centre
                                       // of the mesh (possibly the centre of gravity?)
                                       // (relative coordinates, just like the vertices)
   bit32      CollisionSize;           // radius of collisional sphere
   bit16      NumVertices;             // number of vertices in this mesh
   tr2_vertex *Vertices;               // list of vertices (relative coordinates)
   bit16      NumNormals;              // number of normals in this mesh (should always equal NumVertices)
   tr2_vertex *Normals;                // list of normals (NULL if NumNormals < 0)
   bit16      *MeshLights;             // if NumNormals < 0
   bit16      NumTexturedRectangles;   // number of textured rectangles in this mesh
   tr2_face4  *TexturedRectangles;     // list of textured rectangles
   bit16      NumTexturedTriangles;    // number of textured triangles in this mesh
   tr2_face3  *TexturedTriangles;      // list of textured triangles
   bit16      NumColouredRectangles;   // number of coloured rectangles in this mesh
   tr2_face4  *ColouredRectangles;     // list of coloured rectangles
   bit16      NumColouredTriangles;    // number of coloured triangles in this mesh
   tr2_face3  *ColouredTriangles;      // list of coloured triangles
   } tr2_mesh;

/*
 * Frame structure.
 *
 * Frames indicates how composite meshes are positioned and rotated.  They work
 * in conjunction with Animations[] and Bone2[].  A given frame has the following
 * format:
 *    bit16 BB1x, BB1y, BB1z           // bounding box (low)
 *    bit16 BB2x, BB2y, BB2z           // bounding box (high)
 *    bit16 OffsetX, OffsetY, OffsetZ  // starting offset for this moveable
 *    (TR1 ONLY: bit16 NumValues       // number of angle sets to follow)
 *    (TR2/3: NumValues is implicitly NumMeshes (from moveable))
 *    What follows next is a list of angle sets.  In TR2/3, an angle set can
 *    specify either one or three axes of rotation.  If either of the high two
 *    bits (0xc000) of the first angle bitu16 are set, it's one axis:  only one
 *    bitu16, low 10 bits (0x03ff), scale is 0x100 == 90 degrees;  the high two
 *    bits are interpreted as follows:  0x4000 == X only, 0x8000 == Y only,
 *    0xC000 == Z only.
 *    If neither of the high bits are set, it's a three-axis rotation.  The next
 *    10 bits (0x3ff0) are the X rotation, the next 10 (including the following
 *    bitu16) (0x000f, 0xfc00) are the Y rotation, the next 10 (0x03ff) are the
 *    Z rotation, same scale as before (0x100 == 90 degrees).
 *    Rotations are performed in Y, X, Z order.
 *    TR1 ONLY: All angle sets are two words and interpreted like the two-word
 *    sets in TR2/3, EXCEPT that the word order is reversed.
 */
typedef struct tr2_frame_struct {
   tr2_vertex Vector[3];
   int        NumWords;
   bitu16     *Words;
   } tr2_frame;

/*
 * Item structure
 */
typedef struct tr2_item_struct { // 24 bytes
   bit16 ObjectID;
   bit16 Room;
   bit32 x;
   bit32 y;
   bit32 z;
   bit16 Angle;
   bit16 Intensity1;
   bit16 Intensity2;
   bit16 Flags;   // 0x0100 indicates "inactive" or "invisible"
   } tr2_item;

/*
 * SoundSource structure
 */
typedef struct tr2_sound_source_struct {
   bit32  x,               // position of sound source
          y,
          z;
   bitu16 SoundID;         // internal sound index
   bitu16 Flags;           // 0x40, 0x80, or 0xc0
   bitu8  junk[24];        // TR4 only
   } tr2_sound_source;

/*
 * Boxes structure
 */
typedef struct tr2_box_struct {
   bitu8 Zmin,             // sectors (* 1024 units)
         Zmax,
         Xmin,
         Xmax;
   bit16 TrueFloor;        // Y value (no scaling)
   bit16 OverlapIndex;     // index into Overlaps[]
   } tr2_box;

/*
 * AnimatedTexture structure
 * - really should be simple bit16[], since it's variable length
 */
typedef struct tr2_animated_texture_struct {
   bit16 NumTextureIDs;    // Number of Texture IDs - 1
   bit16 *TextureList;     // list of textures to cycle through
   } tr2_animated_texture;

/*
 * Camera structure
 */
typedef struct tr2_camera_struct {
   bit32 x,
         y,
         z;
   bit16 Room;
   bitu16 Unknown1;  // correlates to Boxes[]?
   } tr2_camera;

/*
 * Sound sample structure
 */
typedef struct tr2_sound_details_struct {   // 8 bytes
   bit16 Sample;
   bit16 Volume;
   bit16 SoundRange;
   bit16 Flags;            // bits 8-15: priority?, 2-7: number of sound samples
                           // in this group, bits 0-1: channel number
   } tr2_sound_details;

/*
 * Cutscene Camera structure
 */
typedef struct tr2_cinematic_frame_struct {
   bit16 rotY,             // Rotation about Y axis, +/-32767 ::= +/- 180 degrees
         rotZ,             // Rotation about Z axis, +/-32767 ::= +/- 180 degrees
         rotZ2,            // Rotation about Z axis (why two?), +/-32767 ::= +/- 180 degrees
         posZ,             // Z position of camera, relative to something
         posY,             // Y position of camera, relative to something
         posX,             // X position of camera, relative to something
         Unknown1,
         rotX;             // Rotation about X axis, +/-32767 ::= +/- 180 degrees
   bitu8 filler[8];        // TR4 only
   } tr2_cinematic_frame;

/*
 * Structure to contain an entire level.
 */
typedef struct tr2_level_struct {
   char   *FileName;                            // filename (not in .TR2 file)
   bitu32 Version;                              // .TR2 file version
   TR2_Version_type EngineVersion;              // TombRaider_1, TombRaider_2, TombRaider_3
   struct tr2_colour_struct Palette8[256];      // 8-bit palette
   bitu32 Palette16[256];                       // 16-bit palette
   bitu32 NumTextiles;                          // number of textiles
   struct tr2_textile8_struct *Textile8;        // 8-bit (palettised) textiles
   struct tr2_textile16_struct *Textile16;      // 16-bit (ARGB) textiles
   bitu32 UnknownT;                             // 32-bit unknown (always 0 in real TR2 levels)
   bitu16 NumRooms;                             // number of rooms in this level
   struct tr2_room_struct *Rooms;               // list of rooms
   bitu32 NumFloorData;                         // number of words of floor data this level
   bitu16 *FloorData;                           // floor data
   int    NumMeshes;                            // number of meshes this level
   struct tr2_mesh_struct *Meshes;              // list of meshes
   bitu32 NumAnimations;                        // number of animations this level
   struct tr2_animation_struct *Animations;     // list of animations
   bitu32 NumStateChanges;                      // number of structures(?) this level
   struct tr2_state_change_struct *StateChanges; // list of structures
   bitu32 NumAnimDispatches;                    // number of ranges(?) this level
   struct tr2_anim_dispatch_struct *AnimDispatches;    // list of ranges
   bitu32 NumAnimCommands;                      // number of Bone1s this level
   struct tr2_anim_command_struct *AnimCommands; // list of Bone1s
   bitu32 NumMeshTrees;                         // number of Bone2s this level
   struct tr2_meshtree_struct *MeshTrees;       // list of Bone2s
   bitu32 NumFrames;                            // number of words of frame data this level
   bitu16 *Frames;                              // frame data
   bitu32 NumMoveables;                         // number of moveables this level
   struct tr2_moveable_struct *Moveables;       // list of moveables
   bitu32 NumStaticMeshes;                      // number of static meshes this level
   tr2_staticmesh *StaticMeshes;                // static meshes
   bitu32 NumObjectTextures;                    // number of object textures this level
   struct tr2_object_texture_struct *ObjectTextures;  // list of object textures
   bitu32 NumSpriteTextures;                    // number of sprite textures this level
   struct tr2_sprite_texture_struct *SpriteTextures; // list of sprite textures
   bitu32 NumSpriteSequences;                   // number of sprite sequences this level
   struct tr2_sprite_sequence_struct *SpriteSequences; // sprite sequence data
   bit32  NumCameras;                           // Number of Cameras
   tr2_camera  *Cameras;                        // cameras
   bit32  NumSoundSources;                      // Number of Sounds
   tr2_sound_source *SoundSources;              // sounds
   bit32  NumBoxes;                             // Number of Boxes
   tr2_box *Boxes;                              // boxes - looks like struct { bitu16 value[4]; } - value[0..2] might be a vector; value[3] seems to be index into Overlaps[]
   bit32  NumOverlaps;                          // Number of Overlaps
   bit16  *Overlaps;                            // overlaps - looks like bitu16; 0x8000 is flag of some sort
                                                //            appears to be an offset into Boxes[] and/or Boxes2[]
   bit16  *Zones;                               // boxes2
   bit32  NumAnimatedTextures;                  // Number of AnimTextures
   bit16  *AnimatedTextures;                    // animtextures
   bit32  NumItems;                             // Number of Items
   struct tr2_item_struct *Items;               // items
   bitu8  *LightMap;                            // colour-light maps
   bitu16 NumCinematicFrames;                   // number of cut-scene frames
   tr2_cinematic_frame *CinematicFrames;        // cut-scene frames
   bit16  NumDemoData;                          // Number of Demo Data
   bitu8  *DemoData;                            // demo data
   bit16  *SoundMap;                            // sound map
   bit32  NumSoundDetails;                      // Number of SampleModifiers
   struct tr2_sound_details_struct *SoundDetails; // sample modifiers
   bit32  NumSampleIndices;                     // Number of Sample Indices
   bit32  *SampleIndices;                       // sample indices
   } tr2_level;

typedef tr2_level TR2_Level;

/*
 * Function Prototypes
 */
#ifndef OPENGL_CPP
extern TR2_Level *ReadTR2level(char *FileName);
extern void FreeTR2level(TR2_Level *Level);
extern int WriteTR2level(TR2_Level *Level, char *FileName);
extern int DumpTR2level(TR2_Level *Level, char *FileName);
#endif

extern void TR2fmake(tr2_file *fp, TR2_Version_type version,
                     FILE *ffp, bitu8 *data, bitu32 datasize);
extern int TR2fread(void *buffer, size_t size, size_t count, tr2_file *fp);
extern int TR2fwrite(void *buffer, size_t size, size_t count, tr2_file *fp);
extern int TR2fseek(tr2_file *fp, long offset, int whence);
extern long TR2fcopy(tr2_file *fp, long offset, tr2_file *gp);

extern int TR2fread1_bitu8(bitu8 *buf, tr2_file *fp);
extern int TR2fread_bitu8(bitu8 *buf, size_t count, tr2_file *fp);
extern int TR2fread1_bit8(bit8 *buf, tr2_file *fp);
extern int TR2fread_bit8(bit8 *buf, size_t count, tr2_file *fp);
extern int TR2fread1_bitu16(bitu16 *buf, tr2_file *fp);
extern int TR2fread_bitu16(bitu16 *buf, size_t count, tr2_file *fp);
extern int TR2fread1_bit16(bit16 *buf, tr2_file *fp);
extern int TR2fread_bit16(bit16 *buf, size_t count, tr2_file *fp);
extern int TR2fread1_bitu32(bitu32 *buf, tr2_file *fp);
extern int TR2fread_bitu32(bitu32 *buf, size_t count, tr2_file *fp);
extern int TR2fread1_bit32(bit32 *buf, tr2_file *fp);
extern int TR2fread_bit32(bit32 *buf, size_t count, tr2_file *fp);

extern int TR2fwrite1_bitu8(bitu8 *buf, tr2_file *fp);
extern int TR2fwrite_bitu8(bitu8 *buf, size_t count, tr2_file *fp);
extern int TR2fwrite1_bit8(bit8 *buf, tr2_file *fp);
extern int TR2fwrite_bit8(bit8 *buf, size_t count, tr2_file *fp);
extern int TR2fwrite1_bitu16(bitu16 *buf, tr2_file *fp);
extern int TR2fwrite_bitu16(bitu16 *buf, size_t count, tr2_file *fp);
extern int TR2fwrite1_bit16(bit16 *buf, tr2_file *fp);
extern int TR2fwrite_bit16(bit16 *buf, size_t count, tr2_file *fp);
extern int TR2fwrite1_bitu32(bitu32 *buf, tr2_file *fp);
extern int TR2fwrite_bitu32(bitu32 *buf, size_t count, tr2_file *fp);
extern int TR2fwrite1_bit32(bit32 *buf, tr2_file *fp);
extern int TR2fwrite_bit32(bit32 *buf, size_t count, tr2_file *fp);

extern void *TR2alloc(int size);
extern void FreeIfUsed(void *mem);

#endif   // #ifdef TR2_H


