
/*
 * tr2io.c
 *
 * Tomb Raider II input/output routines for TRueView
 *
 *** CURRENTLY LACKING:
 ***     Better error control (specifically, any malloc() or fread() error
 ***                           just unceremoniously exits the program)
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <string.h>

#include "config.h"
#include "tr2port.h"
#include "tr2.h"

/***
 *** Code and Macros to handle Big Endian and Little Endian systems
 ***
 *** Users with big-endian CPUs (e.g. Mac) need to
 *** #define BIG_ENDIAN_CPU before this point (see TR2PORT.H).
 ***/

#ifdef BIG_ENDIAN_CPU   // don't include this code if we're on an Intel CPU

/*
 * Swap16(bitu16 *value)
 *
 * Inverts the bytes of a 16-bit value (to handle big-endian CPUs)
 */
static void Swap16(bitu16 *value)
{
   union {
      bitu16 u16;
      bitu8  u8[2];
      } swap;
   bitu8 ch;

   swap.u16 = *value;
   ch = swap.u8[0];
   swap.u8[0] = swap.u8[1];
   swap.u8[1] = ch;
   *value = swap.u16;
}

/*
 * Swap32(bitu32 *value)
 *
 * Inverts the bytes of a 32-bit value (to handle big-endian CPUs)
 */
static void Swap32(bitu32 *value)
{
   union {
      bitu32 u32;
      bitu8  u8[4];
      } swap;
   bitu8 ch;

   swap.u32 = *value;
   ch = swap.u8[0];
   swap.u8[0] = swap.u8[3];
   swap.u8[3] = ch;
   ch = swap.u8[1];
   swap.u8[1] = swap.u8[2];
   swap.u8[2] = ch;
   *value = swap.u32;
}

/*
 * Generic byte swapper (calls Swap16 or Swap32 based on <size>)
 */
static void Swap16_32(bitu32 *value, int size)
{
   if (size == sizeof(bitu16))
      Swap16((bitu16 *)value);
   else // size == sizeof(bitu32)
      Swap32(value);
}

#endif   // BIG_ENDIAN_CPU

/*
 * Macros for handling endian-ness.
 * Note that if we're on a LITTLE_ENDIAN_CPU (e.g. Intel), this expands
 * to nothing (thus introducing no overhead).
 * Mac users need to #define BIG_ENDIAN_CPU so that all integral values
 * longer than 8 bits are handled correctly.
 */
#ifdef   BIG_ENDIAN_CPU
#define  CONVERT_ENDIAN(x, size) Swap16_32((bitu32 *)(x), (size))
#else    // LITTLE_ENDIAN_CPU (e.g. Intel) (default - don't do anything)
#define  CONVERT_ENDIAN(x, size)
#endif   // BIG_ENDIAN_CPU

/*
 * TR2alloc()
 *
 * Wrapper for malloc() - if malloc() returns NULL, we can handle
 * it here (all I do right now is exit).
 */
void *TR2alloc(int size)
{
   void *ReturnValue;

   if (size == 0)
      return(NULL);
//
// I've been told that MacOS has problems here, and that doubling
// the allocation size (e.g. ReturnValue = malloc(size * 2)) "fixes"
// the problem (or at least masks it well)...
//
#ifdef MACINTOSHPC
   size *= 2;
#endif
   if ((ReturnValue = malloc(size)) == NULL) {
      fprintf(stderr, "malloc(%d) failed, exiting...\n", size);
      size = 1 / (int)ReturnValue;  // do a divide-by-zero so we can get into the debugger here
      exit(1);
      }
   memset(ReturnValue, 0x00, size); // initialise it to NULL (better safe than sorry)
   return(ReturnValue);
}

/*
 * AllocTR2level()
 *
 * Allocates a TR2_level base structure.
 */

TR2_Level *
AllocTR2level( void)
{
   TR2_Level *Level;

   Level = (TR2_Level *)TR2alloc(sizeof(TR2_Level));
   return(Level);
}

/*
 * FreeIfUsed()
 *
 * Frees a block of memory if the pointer is non-NULL.
 * (worker-bee for FreeTR2level)
 */
void FreeIfUsed(void *mem)
{
   if (mem != NULL)
      free(mem);
}

/*
 * FreeTR2level()
 *
 * Frees a TR2_Level structure in its entirety, including the
 * base structure.
 */
void FreeTR2level(TR2_Level *Level)
{
   int i;

   FreeIfUsed(Level->FileName);
   FreeIfUsed(Level->Textile8);
   FreeIfUsed(Level->Textile16);

   /* Rooms have internal allocated structures...*/
   for (i = 0; i < Level->NumRooms; ++i) {
      FreeIfUsed(Level->Rooms[i].RoomData.Vertices);
      FreeIfUsed(Level->Rooms[i].RoomData.Rectangles);
      FreeIfUsed(Level->Rooms[i].RoomData.Triangles);
      FreeIfUsed(Level->Rooms[i].RoomData.Sprites);
      FreeIfUsed(Level->Rooms[i].Data);   // this _should_ already be free (NULL)
      FreeIfUsed(Level->Rooms[i].Portals);
      FreeIfUsed(Level->Rooms[i].SectorList);
      FreeIfUsed(Level->Rooms[i].Lights);
      FreeIfUsed(Level->Rooms[i].StaticMeshes);
      }
   FreeIfUsed(Level->Rooms);

   FreeIfUsed(Level->FloorData);

   /* Meshes have internal allocated structures...*/
   for (i = 0; i < Level->NumMeshes; ++i) {
      FreeIfUsed(Level->Meshes[i].Vertices);
      FreeIfUsed(Level->Meshes[i].Normals);
      FreeIfUsed(Level->Meshes[i].MeshLights);
      FreeIfUsed(Level->Meshes[i].TexturedRectangles);
      FreeIfUsed(Level->Meshes[i].TexturedTriangles);
      FreeIfUsed(Level->Meshes[i].ColouredRectangles);
      FreeIfUsed(Level->Meshes[i].ColouredTriangles);
      }
   FreeIfUsed(Level->Meshes);

   FreeIfUsed(Level->Animations);
   FreeIfUsed(Level->StateChanges);
   FreeIfUsed(Level->AnimDispatches);
   FreeIfUsed(Level->AnimCommands);
   FreeIfUsed(Level->MeshTrees);
   FreeIfUsed(Level->Frames);
   FreeIfUsed(Level->Moveables);
   FreeIfUsed(Level->StaticMeshes);
   FreeIfUsed(Level->ObjectTextures);
   FreeIfUsed(Level->SpriteTextures);
   FreeIfUsed(Level->SpriteSequences);
   FreeIfUsed(Level->Cameras);
   FreeIfUsed(Level->SoundSources);
   FreeIfUsed(Level->Boxes);
   FreeIfUsed(Level->Overlaps);
   FreeIfUsed(Level->Zones);
   FreeIfUsed(Level->AnimatedTextures);
   FreeIfUsed(Level->Items);
   FreeIfUsed(Level->LightMap);
   FreeIfUsed(Level->CinematicFrames);
   FreeIfUsed(Level->DemoData);
   FreeIfUsed(Level->SoundMap);
   FreeIfUsed(Level->SoundDetails);
   FreeIfUsed(Level->SampleIndices);

   FreeIfUsed(Level);
}

/*
 * TR2fmake
 */

void
TR2fmake(tr2_file *fp, TR2_Version_type version, 
         FILE *ffp, bitu8 *data, bitu32 datasize)
{
   fp->Version = version; fp->Fp = ffp;
   fp->Data = data; fp->DataSize = datasize; fp->DataOffset = 0;
   return;
}

/*
 * TR2ftell
 */

long
TR2ftell(tr2_file *fp)
{
  long offset;

  if (fp->Fp != NULL) {
    offset = ftell(fp->Fp);
    if (offset < 0) {
      fprintf(stderr, "ftell(file) returned %ld\n", offset);
      exit(2);
      }
    }
  else {
    offset = (long) fp->DataOffset;
    }

  return(offset);
}

/*
 * TR2fseek
 */

int
TR2fseek(tr2_file *fp, long offset, int whence)
{
   int rc = 0;
   bitu32 pos;

   if (fp->Fp != NULL) {
      rc = fseek(fp->Fp, offset, whence);
      if (rc != 0) {
         fprintf(stderr, "fseek(file, %ld, %d) returned %d\n",
            offset, whence, rc);
         exit(2);
         }
      }
   else {
      if (whence == SEEK_SET) pos = 0;
      else if (whence == SEEK_CUR) pos = fp->DataOffset;
      else if (whence == SEEK_END) pos = fp->DataSize;
      else {
         fprintf(stderr, "fseek(data, %ld, %d) failed\n", offset, whence);
         exit(2);
         }
      pos += offset;
      if ((pos < 0) || (pos > fp->DataSize)) {
         fprintf(stderr, "fseek(data, %ld, %d) failed\n", offset, whence);
         exit(2);
         }
      fp->DataOffset = pos;
      }

   return(rc);
}

/*
 * TR2fread()
 *
 * Wrapper for fread() - if fread() fails to return the number of
 * items requested, print a message and exit.
 */

int
TR2fread(void *buffer, size_t size, size_t count, tr2_file *fp)
{
   size_t items;
   int NumRead = (int) (size * count);

   if (fp->Fp != NULL) {
      items = fread(buffer, size, count, fp->Fp);
      if (items != count) {
         fprintf(stderr, "fread(buffer, %d, %d, file) returned %d\n",
            size, count, items);
         exit(2);
         }
      }
   else {
      if ((fp->DataOffset + NumRead) <= fp->DataSize) {
         memcpy(buffer, fp->Data + fp->DataOffset, NumRead);
         fp->DataOffset += NumRead;
         }
      else {
         fprintf(stderr, "fread(buffer, %d, %d, data) failed\n", size, count);
         exit(2);
         }
      }
   return(NumRead);
}

/*
 * Auxiliary functions for reading basic data, including conversion.
 */

int
TR2fread1_bitu8(
  bitu8 *buf, tr2_file *fp)
{
  int NumRead;

  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
  return(NumRead);
}

int
TR2fread_bitu8(
  bitu8 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
  return(NumRead);
}

int
TR2fread1_bit8(
  bit8 *buf, tr2_file *fp)
{
  int NumRead;
 
  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
  return(NumRead);
}
 
int
TR2fread_bit8(
  bit8 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
  return(NumRead);
}

int
TR2fread1_bitu16(
  bitu16 *buf, tr2_file *fp)
{
  int NumRead;

  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}

int
TR2fread_bitu16(
  bitu16 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}

int
TR2fread1_bit16(
  bit16 *buf, tr2_file *fp)
{
  int NumRead;
 
  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}
 
int
TR2fread_bit16(
  bit16 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}

int
TR2fread1_bitu32(
  bitu32 *buf, tr2_file *fp)
{
  int NumRead;

  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}

int
TR2fread_bitu32(
  bitu32 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}

int
TR2fread1_bit32(
  bit32 *buf, tr2_file *fp)
{
  int NumRead;
 
  NumRead = TR2fread(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
}
 
int
TR2fread_bit32(
  bit32 *buf, size_t count, tr2_file *fp)
{
  int NumRead;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
  NumRead = TR2fread(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumRead);
} 

/*
 * TR2fwrite()
 *
 * Wrapper for fwrite() - if fwrite() fails to return the number of
 * items requested, print a message and exit.
 */
int
TR2fwrite(void *buffer, size_t size, size_t count, tr2_file *fp)
{
   size_t items;
   int NumWrite = (int) (size * count);

   if (fp->Fp != NULL) {
      items = fwrite(buffer, size, count, fp->Fp);
      if (items != count) {
         perror("TR2fwrite");
         fprintf(stderr, "fwrite(buffer, %d, %d, file) returned %d\n",
            size, count, items);
         exit(3);
         }
      }
   else {
      if ((fp->DataOffset + NumWrite) <= fp->DataSize) {
         memcpy(fp->Data + fp->DataOffset, buffer, NumWrite);
         fp->DataOffset += NumWrite;
         }
      else {
         fprintf(stderr, "fwrite(buffer, %d, %d, data) failed\n", size, count);
         exit(3);
         }
      }
   return(NumWrite);
}

/*
 * Auxiliary functions for writing basic data, including conversion.
 *
 * Note that endian-ness conversion takes place twice: once to fix the values
 * before writing, and once to put them back again afterward (assuming you
 * want to keep manipulating the same level after a save).
 */

int
TR2fwrite1_bitu8(
  bitu8 *buf, tr2_file *fp)
{
  int NumWrite;

  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
  return(NumWrite);
}

int
TR2fwrite_bitu8(
  bitu8 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
 
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
  return(NumWrite);
}

int
TR2fwrite1_bit8(
  bit8 *buf, tr2_file *fp)
{
  int NumWrite;
 
  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
  return(NumWrite);
}
 
int
TR2fwrite_bit8(
  bit8 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
 
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
  return(NumWrite);
}

int
TR2fwrite1_bitu16(
  bitu16 *buf, tr2_file *fp)
{
  int NumWrite;

#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}

int
TR2fwrite_bitu16(
  bitu16 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}

int
TR2fwrite1_bit16(
  bit16 *buf, tr2_file *fp)
{
  int NumWrite;
 
#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap16(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}
 
int
TR2fwrite_bit16(
  bit16 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap16(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}

int
TR2fwrite1_bitu32(
  bitu32 *buf, tr2_file *fp)
{
  int NumWrite;

#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}

int
TR2fwrite_bitu32(
  bitu32 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}

int
TR2fwrite1_bit32(
  bit32 *buf, tr2_file *fp)
{
  int NumWrite;
 
#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), 1, fp);
#ifdef BIG_ENDIAN_CPU
  Swap32(buf);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
}
 
int
TR2fwrite_bit32(
  bit32 *buf, size_t count, tr2_file *fp)
{
  int NumWrite;
#ifdef BIG_ENDIAN_CPU
  int i;
#endif  // BIG_ENDIAN_CPU
 
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  NumWrite = TR2fwrite(buf, sizeof(*buf), count, fp);
#ifdef BIG_ENDIAN_CPU
  for (i = 0; i < count; i++)
    Swap32(buf + i);
#endif  // BIG_ENDIAN_CPU
  return(NumWrite);
} 

/*
 * TR2fcopy
 */

long
TR2fcopy(
  tr2_file *fp, long offset, tr2_file *gp)
{
  long curoff;
  size_t size;
  void *buf, *data;
 
  /* save current position */
  curoff = TR2ftell(fp);
  if (offset >= curoff)
    return(curoff);

  /* read data */
  size = curoff - offset;
  TR2fseek(fp, offset, SEEK_SET);
  buf = TR2alloc(size);
  TR2fread(buf, size, 1, fp);
  TR2fseek(fp, curoff, SEEK_SET);  // safety
 
  /* extend buffer */
  if ((gp->Fp == NULL) && (gp->DataSize < gp->DataOffset + size)) {
    data = TR2alloc(gp->DataOffset + size);
    if (gp->Data != NULL) memcpy(data, gp->Data, gp->DataOffset);
    FreeIfUsed(gp->Data); gp->Data = data;
    gp->DataSize = gp->DataOffset + size;
  }

  /* write data */
  TR2fwrite(buf, size, 1, gp);
  FreeIfUsed(buf);

  return(curoff);
} 

